/*
 * @(#)ecTypes.h	1.5  4/27/88
 */
/* Definitions of Emerald kernel/compiler data structures.   		  *
 * Origin: 1986-09-04							  *
 * Copyright 1986 Eric Jul and Norm Hutchinson.	 May not be used for any  *
 * purpose without written permission from the authors.			  */

#ifndef ecTypes_h
#define ecTypes_h

#define EMVERSION 0x2120addd

/*
 * All object data areas and  object descriptors have an ODTag
 * which includes a tag indicating what the area is:
 */

typedef enum {
    InvalidTag,
    UnknownTag,
    GODTag,
    GODataTag,
    SSODTag,
    SSTag,
    CodeODTag,
    CodeTag,
    LOTag,
    ProcessODTag,
    AbConTag,
    CondTag,
    DotoTag,
    LastODBasicTag
} ODBasicTag;


typedef struct ODTag {
    Boolean	        frozen: 1;	/* 1 => ask the kernel */
    Boolean	        global: 1;	/* 1 => this is an OD with a data 
					   pointer. This is redundant since
					   the basic tag encodes this info. */
    Boolean	        replicated: 1;	/* 1 => may be replicated */
    Boolean		isResident: 1;
    Boolean		setUpDone : 1;
    Boolean             isFixed: 1;
    Boolean             inTransit: 1;

    Boolean		allInstancesAreLocal: 1;
    Boolean		hasNoPointer: 1;

    Boolean             broken: 1;

    Boolean             gcDoNotCollect: 1;
    Boolean	        gcMark1: 1;
    Boolean	        gcMark2: 1;
    Boolean             gcFrozen: 1;
    Boolean             localgcMark1: 1;
    Boolean             localgcMark2: 1;
    Boolean             localgcFrozen: 1;

    Boolean             dormant:1;
    Boolean             stopped:1;
    
    Boolean             seenHere:1;
    
    unsigned int        otherstuff: 4;
    ODBasicTag		tag: 8;
} ODTag;

/*
 * otherStuff is to be set to ObscureValue for each ODTag.
 */

#define OBSCUREVALUE 0x5

typedef struct InvokeQueue {
    struct InvokeQueue	*next;
    struct InvokeQueue	*prev;
    struct StackSegment *mySSPtr;
} InvokeQueue, *InvokeQueuePtr;

/*
 * Vectors:
 * Vectors can only be handled with static templates.  There is a Brand,
 * VectorBrand.  The element brand of the vector is described by using the 
 * length field to encode another brand.  Note that the VarVectorBrand may
 * never be used.  A vector contains as its first thing a 32-bit integer 
 * giving the number of entries in the vector.  If the elementBrand is
 * DataBrand, then the length is of course in bytes.
 */

/*
 * Note on the definition of enumerated types:
 * These are often defined twice, once as typedef enum, once as
 * #define mumble 17.   This is due to an error in the cc compiler:
 * when extracting enum values from bit field, the result is sign extended
 * instead of being zero extended -- hence the typedef unsigned int.
 */

typedef unsigned int Brand;
#define DataBrand	    0 /* Raw bits (in units of 8 bit bytes). */
#define ODPBrand	    1 /* ODPs */
#define AddrBrand	    2 /* Machine Address inside somewhere */
#define VectorBrand	    3 /* Vector, next brand describes the elements */
#define VariableBrand	    4 /* variables; both an abconptr and an ODP */
#define MonitorBrand	    5 /* Monitors - 4 byte locked bit, 4 byte queue */
#define InvokeQueueBrand    6 /* Invoke queue, 8 bytes */
#define UnusedBrand	    7

typedef unsigned int TEFormat;
#define ShortDynamicF	0	/* A short, single word entry */
#define LongDynamicF	1	/* A long entry */
#define ShortStaticF	2	/* Static entry */
#define RegisterF	3	/* Register entry */

/* Currently, dynamics are not used;  if they ever are,
   they should have attached flags too. */
/*  */
/* typedef struct ShortDynamic { */
/*     TEFormat		Format          : 2; */
/*     Brand		theBrand        : 3; */
/*     Offset		dataOffset      : 12; */
/*     Offset		codeOffset      : 15; */
/* } ShortDynamic; */
/*  */
/* typedef struct LongDynamic { */
/*     TEFormat		Format          : 2; */
/*     Brand		theBrand        : 3; */
/*     Offset		dataOffset      : 27; */
/*     Offset		codeOffset; */
/* } LongDynamic; */

typedef unsigned int ParamInfo;

#define IsNotParam          0
#define IsArgument          1
#define IsResult            2
#define IsMoveResult        3

typedef struct ShortStatic {
    TEFormat		Format 		: 2;
    Brand		theBrand	: 3;
    Boolean             attachedFlag    : 1;
    ParamInfo           paramInfo       : 2;
    Brand		elementBrand	: 3;
    Offset		count		: 21;
} ShortStatic;

typedef enum {
    InRegisterx,	/* The register described is in a CPU register */
    InSaveAreax		/* The register described is in a save area. */
} Placementx;

typedef unsigned int Placement;
#define InRegister 0
#define InSaveArea 1

typedef struct RegisterTE {
    TEFormat		Format 		: 2;
    Brand		theBrand 	: 3;
    Placement		storedWhere 	: 1;
    Boolean             attachedFlag    : 1;
    Boolean		unusedFlag	: 1;
    unsigned int	reg		: 8;
    unsigned int	count		: 16;
} RegisterTE;


typedef struct TemplateEntry {
    union {
/*	ShortDynamic	SD; */
/*	LongDynamic	LD; */
	ShortStatic	SS;
	RegisterTE	R;
    } TE;
} TemplateEntry, *TemplateEntryPtr;

typedef struct TemplateBits {
    unsigned int	numEntries:24;
    Boolean             hasInvokeQueue:1;
    Boolean             exclusiveAccess:1;
    unsigned int        notUsed:6;
} TemplateBits;

typedef struct Template {
    TemplateBits	B;
    TemplateEntry	entry[ /* numEntries */ 12 ];
} Template, *TemplatePtr;

#define NEWTemplate(N) ((TemplatePtr) malloc(\
  (int) (sizeof(Template) -\
  12 * sizeof(TemplateEntry) +\
  (N) * sizeof(TemplateEntry))))

/*
 * Emerald Relocation information
 */

typedef enum {
  AR_OIDToODP,
  AR_OIDToODAP,
  AR_OpNumberToAddress,
  AR_OIDOIDToAbCon,
  AR_OIDToCheatingODP,
  AR_OIDToCodePtr
 } AR_Kind;

typedef struct ARelocation {
    AR_Kind		kind:8;
    Offset		where:24;
    OID			theOID1;
    OID			theOID2;
} ARelocation, *ARelocationPtr;

typedef struct RelocationInfo {
    int			numEntries;
    ARelocation		relocation[/* numEntries */ 4];
} RelocationInfo, *RelocationInfoPtr;

#define NEWRelocationInfo(N) ((RelocationInfoPtr) malloc(\
  (int) (sizeof(RelocationInfo) -\
  4 * sizeof(ARelocation) +\
  (N) * sizeof(ARelocation))))

/*
 * MONITOROFFSETWITHINOBJECT  for the forseeable future, this hack
 * will be used to find the address of an object given the address
 * of its monitorlock.
 * NOTE: At this time (July 1987), this offset is only used for test output
 * and for dumping the condMap data structure.  Thus if monitor locks one day
 * start appearing at other offsets, it should be simple to remove this
 * restriction.
 */
#define MONITOROFFSETWITHINOBJECT 8

#define mODPFromMonLockPtr(monLock) ( (ODP) addOffset((monLock), - MONITOROFFSETWITHINOBJECT))

#define CHECKPOINT_DEFAULT	0
#define CHECKPOINT_TO		1
#define CHECKPOINT_AT		2
#define CHECKPOINT_ATALL	3
#define CHECKPOINT_CONFIRM	8

#endif
