/*
 * @(#)emTypes.h	1.8  1/20/89
 */

/* Definitions of Emerald kernel data structures.   Origin: 1986-02-11    *
 * Copyright 1986 Eric Jul and Norm Hutchinson.	 May not be used for any  *
 * purpose without written permission from the authors.			  */

#ifndef emTypes_h
#define emTypes_h

#include "stdTypes.h"

#ifndef OMAGIC
#include <a.out.h>
#endif	OMAGIC

#ifndef ecTypes_h
#include "ecTypes.h"
#endif  ecTypes_h

#ifndef NULL
#define NULL 0
#endif	NULL

/* Efficient tests for non NULL and NULL */
#define IsNULL(xxx)  (! ( (int) (xxx) ) )
#define NonNULL(xxx) ((int) (xxx) )

/* Offset manipulation */
#define byteOffset(base, ptr) ( (int) ( ( (char *) (ptr)) - (char *) (base) ))
#define addOffset(base, bytecount) ( ((char *) (base)) + (bytecount) )

typedef int                *SSAddr;
typedef int                *CodeAddr;
typedef int                *DataAddr;
typedef int                *GenericPtr;
typedef int                 CodeOffset;
typedef union OD           *ODP;
typedef unsigned long       EmLocation;
typedef unsigned char       Bytes;

/* Nil in the Emerald language is represented as: */
#define EMNIL 0x80000000

/* Efficient tests for non NIL and NIL */
#define IsNIL(xxx)  ( (int) (xxx) == EMNIL )
#define NonNIL(xxx) ( (int) (xxx) != EMNIL )

typedef struct DQueue {
    struct DQueue          *next;
    struct DQueue          *prev;
} DQueue, *DQueuePtr;

/*
 * Emerald locations consist of an 8-bit node id and a 24-bit timestamp
 * which is merely the number of moves attempted by the object.
 */
typedef struct EmLocationStruct {
    unsigned int        lnn:8;
    unsigned int        timestamp:24;
} EmLocationStruct, *EmLocationStructPtr;

#define mGetLocNodeNum(XX)   (((EmLocationStructPtr)(&(XX)))->lnn)
#define mGetLocTimeStamp(XX) (((EmLocationStructPtr)(&(XX)))->timestamp)
#define LOCATIONNEWER(A,B) (mGetLocTimeStamp(A) > mGetLocTimeStamp(B))
#ifdef vax
#define mMakeLocation(theNode, theTimeStamp) \
   ( (EmLocation) (((theTimeStamp) << 8 ) | ((theNode) & 0xFF)))
#define mMakeNodeOIDFromLoc(theLoc) (((theLoc) & 0xFF) << 24)
#define mMakeNodeOIDFromLNN(theLNN) (((theLNN) & 0xFF) << 24)
#endif
#ifdef sun
/* allocated from other end. */
#define mMakeLocation(theNode, theTimeStamp) \
   ( (EmLocation) (((theNode) << 24 ) | ((theTimeStamp) & 0xFFFFFF)))
#define mMakeNodeOIDFromLoc(theLoc) ((theLoc) & 0xFF000000)
#define mMakeNodeOIDFromLNN(theLNN) (((theLNN) & 0xFF) << 24)
#endif

/*
 * Objects with tag LOTag contain their data right after this field,
 * if you like, in the object table entry itself.  Replicated objects look
 * like local objects, except that they have an OID field right after the
 * codePtr, but the Kernel needs to know that they are
 * replicated in order to do the right thing on move and garbage collection.
 */

typedef struct DynamicLink {
    CodeAddr		ip;		/* Instruction ptr */
    struct GOD	       *b;		/* ODP for current global object */
    struct GOData      *g;		/* Current object data area */
    SSAddr		l;		/* Local activation record ptr */
} DynamicLink, *DynamicLinkPtr;

/* The following macro is very dependent on the layout of an activation
 * record.  It depends on where the local activation register, l, points
 * to in relation to the dynamic link.
 */

#define LOFFSETFROMDYNAMICLINK (sizeof(CodeAddr))

/* The following attempts to describe an activation record.
 * Since contains variable sized fields, it cannot be described properly.
 * This suffices since the kernel only really needs to use the link
 * fields.
 * Note also that stacks grow from high addresses toward zero.
 */

typedef struct ActRecord {
 /* Variable		array [ ... ] of AVariable */
 /* Register Save Area  array [ ... ] of AVariable */
 /* Temporaries		array [ ... ] of AVariable */
    InvokeQueue		invokeList; /* Only if needed */
    DynamicLink		returnLink;
 /* Arguments		array [ ... ] of AVariable */
 /* Result		array [ ... ] of AVariable */
} ActRecord, *ActRecPtr;


/* l points to the second entry in the dynamic link */
#define mDynLinkPtrFromL(theL) \
    ((DynamicLinkPtr) addOffset(theL, -LOFFSETFROMDYNAMICLINK))

#define mInvokeQueuePtrFromL(theL) \
    ((InvokeQueuePtr) addOffset(theL, \
	-LOFFSETFROMDYNAMICLINK-sizeof(InvokeQueue)))

#define mLFromInvokeQueuePtr(theIQPtr) \
    ((SSAddr) addOffset(theIQPtr, \
	LOFFSETFROMDYNAMICLINK+sizeof(InvokeQueue)))


#define mParamAreaLowAddrFromL(theL) \
    ((SSAddr) addOffset(theL, -LOFFSETFROMDYNAMICLINK+sizeof(DynamicLink)))

#define mStoppedAtEntry(fSSPtr) \
    (((SSAddr) mInvokeQueuePtrFromL((fSSPtr)->regs.l)) < \
    ((SSAddr) ((fSSPtr)->regs.sp)))

#define mRegsStoppedAtEntry(fRegsPtr) \
    (((SSAddr) mInvokeQueuePtrFromL((fRegsPtr)->l)) < \
    ((SSAddr) ((fRegsPtr)->sp)))

/*
 * The following registers are saved on process preemption.
 */

typedef struct RegisterSave {
    int                 arg1, arg2, arg3, r4,   r5,
			r6,   r7,   r8,   r9;
    struct GOD         *b;
    struct GOData      *g;
    SSAddr              l;
    SSAddr              sp;
} RegisterSave, *RegisterSavePtr;

/*
 * The following macros are used for accessing register that are saved
 * in a RegisterSave area.
 */

#define mSetSavedReg(fRegisterSavePtr, fReg, fValue) \
    * ((int *) addOffset((fRegisterSavePtr), ((fReg) - 1) * sizeof(int))) = \
    (int) fValue

#define mGetSavedReg(fRegisterSavePtr, fReg) \
    (* ((int *) addOffset((fRegisterSavePtr), ((fReg) - 1) * sizeof(int))))

/*
 * Object Descriptors:
 * Every object has at least a minimal descriptor consisting of
 * an ODTag, and if not degenerate also the following:
 */

typedef struct GenericOD {      /* Generic object descriptor */
    ODTag		tag;
    struct GOData      *dataPtr;
    OID			ownOID;
    EmLocation		ownLoc;
    /* Now the OD specific fields. */
} GenericOD;

typedef struct GOD {            /* Object table entry - for Emerald globals */
    ODTag		tag;
    struct GOData      *dataPtr;
    OID			ownOID;
    EmLocation		ownLoc;
    /* Now the OD specific fields. */
    InvokeQueue		ARListHead;
    OID			myCodeOID;
    /* Now the cP-added fields: review checkPoint docs for information */
    long		cPSize, cPOffset, cPSequenceNum;
    EmLocation          cPServerLoc;
    char               *Replicant;
} GOD, *GODP;

typedef struct SSOD {           /* OD - for Stack Segments */
    ODTag		tag;
    struct StackSegment*dataPtr;
    OID			ownOID;
    EmLocation		ownLoc;
    OID                 processOID;
} SSOD, *SSODP;

/*
 * Note that the OD for local objects
 * is also the object data area.
 */
typedef struct LOData {         /* OD/Data area for local objects */
    ODTag		tag;
    struct Code	       *myCodePtr;
    Bytes		inlineData[ /* some size or other */ 8 ];
} LOData, *LODataPtr;


/*
 * Replicated.
 */
typedef struct ROData {		/* OD/Data area for replicated objects */
    ODTag		tag;
    struct Code	       *myCodePtr;
    OID			ownOID;
    Bytes		inlineData[ /* some size or other */ 8 ];
} ROData, *RODataPtr;


typedef struct CodeOD {
    ODTag		tag;
    struct Code	       *dataPtr;
    OID			ownOID;
    EmLocation		ownLoc;
    /* Now the OD specific fields. */
    GenericPtr          AbConSetPtr;
    GenericPtr          depCode;
} CodeOD, *CodeODP, ATOD, *ATODP;

/* Monitor and condition variables */

typedef struct MonitorLock {
    Boolean		 isLocked;
    struct StackSegment	*waiting;
} MonitorLock, *MonitorLockPtr;

typedef struct CondOD {         /* OD for Condition objects */
    ODTag		tag;
    struct GOData      *dataPtr;
    OID			ownOID;
    EmLocation		ownLoc;
    /* Now the OD specific fields. */
    InvokeQueue         ARListHead;
    MonitorLock        *theLock;
    struct StackSegment*waiting;
} CondOD, *CondODP;

typedef struct AbConOpVectorEntry {
    CodeAddr            opAddress;
    short int           argumentCount;
    short int           resultCount;
    OID			operationNameOID;
} AbConOpVectorEntry, *AbConOpVectorEntryPtr;

typedef struct AbCon {
    ODTag               tag;
    OID			ATOID;
    OID			restrictOID;
    OID			CodeOID;
    struct Code	       *aPtr;
    struct Code	       *rPtr;
    struct Code	       *cPtr;
    int                 numEntries;
    AbConOpVectorEntry	opVector[/* numEntries */ 1 ];
} AbCon, *AbConPtr;

#define NEWAbCon(N) ((AbConPtr) emalloc(\
  (int) (sizeof(AbCon) -\
  1 * sizeof(AbConOpVectorEntry) +\
  (N) * sizeof(AbConOpVectorEntry))))

typedef union OD {    /* Union of all kinds of ODs */
    GOD			G;
    CodeOD		C;
    ATOD		A;
    struct LOData	L;		/*
					 * Note that local objects are their
					 * own ODs.
					 */
    struct ROData	R;
    SSOD		SS;
    AbCon               AB;
    CondOD              CD;
} OD /* , *ODP -- done above */;

/*
 * Global object data areas look exactly like local object data areas.	Once
 * we have found the node where the object really resides, we do a local call
 * to it.
 */
typedef struct GOData {			/* Global Object data */
    ODTag		tag;
    struct Code	       *myCodePtr;	/* needs translating on move */
/*
 * The rest of this is dependent on the concrete type.
 */
    Bytes		inlineData[ /* some size or other */ 8 ];
} GOData, *GODataPtr;

/*
 * Stack segments
 */
typedef enum {
    SSNotInUse,
    SSAllocated,
    SSRunnable,
    SSCondWait,
    SSMonWait,
    SSReqWait,
    SSInvokeWait,
    SSTimerWait,
    SSInitiallyWait,
    SSMoveWait,
    SSLocateWait,
    SSConformWait,
    SSReadIOWait,
    SSWriteIOWait,
    SSOwnWait,
    SSBroken,

    SSLastRunStatus        /* Must be last enum in SSRunStatus */
} SSRunStatus;

typedef struct SSStatus {
    SSRunStatus         rs;
} SSStatus, *SSStatusPtr;

typedef struct StackSegment {
    ODTag		tag;
    OID			ownOID;
    OID			processOID;
    SSStatus            status;
    SSODP               ownSSODP;
    int                 availStack;
    int			segmentSize;
    SSAddr              splimit;
    SSAddr		endOfSS;
    struct StackSegment*readyQLink;
    GenericPtr          rPtr;
    DQueue              ssDQ;
    RegisterSave	regs;
    Brand               resultBrand;
    GenericPtr          invokePtr;
    /* stack area follows */
} SS, *SSPtr;

/* Queue macros  */
#define setRQLink(fSSPtr, newValue) \
    (fSSPtr)->readyQLink = newValue

#define getRQLink(fSSPtr) ( (fSSPtr)->readyQLink )

/*
 * This one is used for finding the template given an ip value, or for
 * finding the appropriate handler or line number.
 */
typedef struct IPMapEntry {
    CodeOffset      highIPOffset;
    Offset          offset;
} IPMapEntry, *IPMapEntryPtr;

typedef struct IPMap {  /* IP-to-whatever map */
    Offset              highestIPOffset, lowestIPOffset;
    IPMapEntry          mapEntries[ 4 ];
} IPMap, *IPMapPtr;

#define NEWIPMap(N) ((IPMapPtr) malloc(\
  (int) (sizeof(IPMap) -\
  4 * sizeof(IPMapEntry) + \
  (N) * (sizeof(IPMapEntry)))))

typedef struct OpVectorEntry {
    CodeOffset		offset;
    Offset		paramTemplateOffset;
    short int           argumentCount;
    short int           resultCount;
    OID			operationNameOID;
} OpVectorEntry, *OpVectorEntryPtr;

typedef struct OpVector {
    int			numEntries;
    OpVectorEntry	entry[/* numEntries */ 4 ];
} OpVector, *OpVectorPtr;

#define NEWOpVector(N) ((OpVectorPtr) malloc(\
  (int)(sizeof(OpVector) -\
  4 * sizeof(OpVectorEntry) +\
  (N) * sizeof(OpVectorEntry))))

/*
 * Here is what the compiler produces in the data area of the .o file.
 */

typedef struct Code {
    ODTag		tag;
    OID			ownOID;
    OID                 ownAbstractType;
    time_t		emVersion;
    Offset              codeNameOffset;
    int			sizeOfThisCodeObject;
    int			instanceSize;
    ODTag		instanceTag;
    Offset              ODATemplateOffset;
    Offset		templateMapOffset;
    Offset		failureHandlerMapOffset;
    Offset		unavailableHandlerMapOffset;
    Offset              lineNumberMapOffset;
    Offset		opVecOffset;
    Offset		relocationInfoOffset;
    Offset              abstractTypeInfoOffset;
    OpVectorEntry	initially;
    OpVectorEntry	recovery;
    OpVectorEntry	process;
    Offset		sourceFileNameOffset;
    Offset		debugInfoMapOffset;
    /* The rest of this is not directly addressable, but must be calculated
       from the offsets above.
    String              codeName;
    IPMap		IPToTemplateMap;
    IPMap		IPToFailureHandlerMap;
    IPMap		IPToUnavailableHandlerMap;
    IPMap               IPToLineNumberMap;
    OpVector		opVector;
    RelocationInfo	relocationInfo;
    struct Template	templates[1]; */
} Code, *CodePtr, AT, *ATPtr;

typedef struct CreateDirective {
    int			numEntries;
    OID			createOID[/* numEntries */ 4 ];
} CreateDirective, *CreateDirectivePtr;

#define NEWCreateDirective(N) ((CreateDirectivePtr) malloc(\
  (int) (sizeof(CreateDirective) -\
  4 * sizeof(OID) +\
  (N) * sizeof(OID))))

/*
 * A .o file from the compiler consists of a Code object
 * followed by CreateDirectives.  Each createDirective instructs the kernel
 * to create an instance of the immediately previous code object with the
 * given OID.  If the numEntries field is 0, then no instances are to be
 * created.
 * .o files are produced by the assembler (see as(1)) and follow the format
 * in a.out(5).  The Emerald stuff is all contained in the data area;
 * there is no code area.  The layout of the code is given above, see Code.
 * Emerald uses the general relocation information (all pc-relative
 * references are ignored; Emerald only uses absolute addressing since
 * this allows a .o file to be shipped across the net and translated
 * again).  In addition, Emerald  has its own relocation information (for
 * translating OID to ODP, etc., see loadFile.c).
 */

typedef struct DotoFile {
  struct exec		 header;
  char			 theRest[/* ? */ 4];
} DotoFile, *DotoFilePtr;

#define mDotoFilePtrFromCodePtr(theCodePtr) \
    ((DotoFilePtr) addOffset(theCodePtr, - DotoFile_theRest))
#define dotoCodePtr(X) ((CodePtr) &(X)->theRest[0])
#define dotoCreateDirectivePtr(X) ((CreateDirectivePtr) \
  &(X)->theRest[dotoCodePtr(X)->sizeOfThisCodeObject])
#define dotoSymbolPtr(X) ((struct nlist *) \
  &(X)->theRest[N_SYMOFF((X)->header)-sizeof(struct exec)])
#define dotoRelocPtr(X) ((struct relocation_info *) \
  &(X)->theRest[(X)->header.a_data])
#define dotoStringPtr(X) ((char *) \
  &(X)->theRest[N_STROFF((X)->header)-sizeof(struct exec)])

#define isOkDoto(X) ((X)->header.a_magic == OMAGIC && \
  (X)->header.a_text == 0 && \
  (X)->header.a_data > 0 && \
  (X)->header.a_bss == 0 && \
  (X)->header.a_trsize == 0)

typedef struct AVariable {
    DataAddr            myAddr;
    struct AbCon       *myAbConPtr;
} AVariable, *AVariablePtr;

/* Vector area */

typedef struct VectorArea {
    int			count; /* Note, name should be sizeInBytes */
    Bytes		data[ /* count */ 8];
} VectorArea, *VectorAreaPtr;

typedef struct Vector {
    ODTag		 tag;
    CodePtr		 myCodePtr;
    int                  upb;
    int			 sizeInBytes;
    Bytes		 data[ /* sizeInBytes */ 8];
} Vector, *VectorPtr;

typedef struct IMVector {
    ODTag		 tag;
    CodePtr		 myCodePtr;
    OID			 ownOID;
    int                  upb;
    int			 sizeInBytes;
    Bytes		 data[ /* sizeInBytes */ 8];
} IMVector, *IMVectorPtr;

typedef struct String {
    ODTag		 tag;
    CodePtr		 myCodePtr;
    OID			 ownOID;		/* since it is immutable */
    int			 sizeInBytes;
    Bytes                data[ /* numEntries */ 8];
} String, *StringPtr;

/*  External declarations: The following Kernel identifiers are visible */

extern ODP		       *OT;
extern SSPtr                    currentSSP;
extern SSAddr                   splimit;

extern void Create(), InitiallyDone(), InvokeReturn(),
    ProcessDone(), MonitorEntry(), MonitorExit(),
    SignalCond(), WaitCond(), Invoke(), Enqueue(), Dequeue(),
    InsertDQ();

extern DQueuePtr RemoveDQ();

#endif emTypes
/* End emTypes.h */
