/*
 * @(#)typeCheck.c	1.7  1/20/89
 */
#include "assert.h"
#include "error.h"
#include "ident.h"
#include "nodes.h"
#include "builtins.h"
#include "MyParser.h"
#include "sequence.h"
#include "evaluate.h"
#include "map.h"
#include "primitives.h"
#include "flags.h"
#include "buildATs.h"
#include "trace.h"
#include "system.h"
#include "opNames.h"
#include "environment.h"
#include "semantics.h"
#include "typecheck.h"

Boolean checkingTypes = FALSE;
extern Map manifestMap;

extern void newEvaluateManifest();
extern void newAssignTypes();
static void checkAssignmentStatement(), checkPrimitiveStatement();
extern NodePtr evaluateForAT();

/*
 * Type checks job is to ensure that the whole program gets type checked.
 * It does this by knowing what things have children that need to be type
 * checked, and doing them, and nothing else.  All abstract types have been 
 * evaluated at this point.
 */

void doTypeCheck();

extern Map dependManifestMap;
extern void checkDelayedChecks();

void typeCheck(p)
NodePtr p;
{
  NodePtr thing, km;
  checkingTypes = TRUE;
  doTypeCheck(p);
  if (p->tag == P_COMP) {
    Map_For(dependManifestMap, thing, km)
      if (km->b.knowmanifest.isOK) {
	doTypeCheck(km->b.knowmanifest.answer);
      }
    Map_Next
  }
  checkDelayedChecks();
  checkingTypes = FALSE;
}

typedef struct sTCDesc {
  NodePtr		object;
  Symbol		objectSym;
  Boolean		inMonitor		:1;
  unsigned int 		unused			:31;
  struct sTCDesc		*next;
} TCDesc, *TCDescPtr;

static TCDescPtr stack = NULL;

static void TCPush(p)
NodePtr p;
{
  register TCDescPtr t;
  t = (TCDescPtr) calloc(1, sizeof(TCDesc));
  t->object = p;
  t->objectSym = p->b.oblit.name->b.symdef.symbol;
  t->next = stack;
  stack = t;
}

static void TCPop()
{
  TCDescPtr this;

  /* Pop it from the stack. */
  this = stack;
  stack = this->next;
  free((char *)this);
}

void doTypeCheck(p)
NodePtr p;
{
  register NodePtr q, at;
  register Symbol st;
  register int i;

  checkingTypes = TRUE;
  if (ISTOKEN(p)) return;
  switch (p->tag) {
    case P_GLOBALREF:
      resolveGlobal(p, (ValuePtr)NULL);
      doTypeCheck(p->b.globalref.value);
      break;
    case P_OPNAME:
    case P_WHEREWIDGIT:
      return;
    case T_SEQUENCE:
      for (i = p->firstChild; i < p->nChildren; i++)
	doTypeCheck(p->b.children[i]);
      break;
    case P_COMP:
      doTypeCheck(p->b.comp.consts);
      break;
    case P_OBLIT:
      if (! p->b.oblit.f.typesAreAssigned) {
	TRACE1(delay, 1, "typecheck: type not assigned in %s\n", ATName(p));
	newAssignTypes(p, 1);
      }
      if (p->b.oblit.f.typesHaveBeenChecked) return;
      p->b.oblit.f.typesHaveBeenChecked = TRUE;
      TCPush(p);
      doTypeCheck(p->b.oblit.decls);	/* declarations */
      doTypeCheck(p->b.oblit.monitor);	/* monitor */
      doTypeCheck(p->b.oblit.ops);	/* operations */
      doTypeCheck(p->b.oblit.process);	/* process */
      if (p->b.oblit.myat != NN) doTypeCheck(p->b.oblit.myat);
      if (p->b.oblit.instat != NN) doTypeCheck(p->b.oblit.instat);
      if (p->b.oblit.name->b.symdef.symbol->value.ATinfo != NN)
	doTypeCheck(p->b.oblit.name->b.symdef.symbol->value.ATinfo);
      TCPop();
      break;
    case P_ATLIT:
      if (! p->b.oblit.f.typesAreAssigned) {
	TRACE1(delay, 1, "typecheck: type not assigned in %s\n", ATName(p));
	newAssignTypes(p, 1);
      }
      if (p->b.atlit.f.typesHaveBeenChecked) return;
      p->b.atlit.f.typesHaveBeenChecked = TRUE;
      doTypeCheck(p->b.atlit.ops);
      break;
    case P_MONITOR:
      stack->inMonitor = TRUE;
      doTypeCheck(p->b.monitor.decls);
      doTypeCheck(p->b.monitor.ops);
      doTypeCheck(p->b.monitor.init);
      doTypeCheck(p->b.monitor.recovery);
      stack->inMonitor = FALSE;
      break;
    case P_INITDEF:
      doTypeCheck(p->b.initdef.body);
      break;
    case P_RECOVERYDEF:
      doTypeCheck(p->b.recoverydef.body);
      break;
    case P_PROCESSDEF:
      doTypeCheck(p->b.processdef.body);
      break;
    case P_ELSECLAUSE:
      doTypeCheck(p->b.elseclause.stats);
      break;
    case P_LOOPSTAT:
      doTypeCheck(p->b.loopstat.stats);
      break;
    case P_FAILUREHANDLER:
      doTypeCheck(p->b.failurehandler.body);
      break;
    case P_OPDEF:
      doTypeCheck(p->b.opdef.sig);	/* signature */
      doTypeCheck(p->b.opdef.body);	/* block */
      break;
    case P_OPSIG:
      doTypeCheck(p->b.opsig.params);
      doTypeCheck(p->b.opsig.results);
      doTypeCheck(p->b.opsig.where);
      break;
    case P_PARAM:
      break;
    case P_VARDECL:
      if (p->b.vardecl.value != NN) {
	q = p->b.vardecl.sym;
	assert(q->tag == P_SYMDEF);
	st = ST_Fetch(q->b.symdef.symbol);
	at = evaluateForAT(p->b.vardecl.value, FALSE);
	if (at->tag == T_SEQUENCE) {
	  if (Sequence_Length(at) != 1) {
	    ErrorMessage(p->b.vardecl.value,
	      "Wrong number of results, expected 1 found %d",
	      Sequence_Length(at));
	  }
	  at = at->b.children[0];
	}
	checkConforms(at, st->value.ATinfo, p->b.vardecl.value);
      }
      break;
    case P_CONSTDECL:
      q = p->b.constdecl.sym;
      assert(q->tag == P_SYMDEF);
      st = ST_Fetch(q->b.symdef.symbol);
      if (st->isManifest || st->hasValue) {
	doTypeCheck(st->value.value);
      } else {
	at = evaluateForAT(p->b.constdecl.value, FALSE);
	if (p->b.constdecl.type != NN) {
	  if (at->tag == T_SEQUENCE) {
	    if (Sequence_Length(at) != 1) {
	      ErrorMessage(p->b.constdecl.value,
		"Wrong number of results, expected 1 found %d",
		Sequence_Length(at));
	    }
	    at = at->b.children[0];
	  }
	  checkConforms(at, st->value.ATinfo, p->b.constdecl.value);
	}
      }
      break;
    case P_IFCLAUSE:
      q = p->b.ifclause.exp;
      at = evaluateForAT(q, TRUE);
      if (! conforms(at, refToBuiltin(B_INSTAT, BOOLEANINDEX))) {
	ErrorMessage(q, "Boolean expression required");
      }
      doTypeCheck(p->b.ifclause.stats);
      break;
    case P_IFSTAT:
      doTypeCheck(p->b.ifstat.ifclauses);
      doTypeCheck(p->b.ifstat.elseclause);
      break;
    case P_EXITSTAT:
      q = p->b.exitstat.exp;
      if (q != NULL) {
	at = evaluateForAT(q, TRUE);
	if (! conforms(at, refToBuiltin(B_INSTAT, BOOLEANINDEX))) {
	  ErrorMessage(q, "Boolean expression required");
	}
      }
      break;
    case P_ASSIGNSTAT:
      checkAssignmentStatement(p);
      break;
    case P_ASSERTSTAT:
      q = p->b.assertstat.exp;
      at = evaluateForAT(q, TRUE);
      if (! conforms(at, refToBuiltin(B_INSTAT, BOOLEANINDEX))) {
	ErrorMessage(q, "Boolean expression required");
      }
      break;
    case P_FIXSTAT:
      at = evaluateForAT(p->b.fixstat.exp, TRUE);	/* just to type check them */
      at = evaluateForAT(p->b.fixstat.loc, TRUE);
      break;
    case P_REFIXSTAT:
      at = evaluateForAT(p->b.refixstat.exp, TRUE);	/* just to type check them */
      at = evaluateForAT(p->b.refixstat.loc, TRUE);
      break;
    case P_UNFIXSTAT:
      at = evaluateForAT(p->b.unfixstat.exp, TRUE);
      break;
    case P_CHECKPOINTSTAT:
    case P_RETURNSTAT:
    case P_RETURNANDFAILSTAT:
      break;
    case P_MOVESTAT:
      /* just to type check them */
      at = evaluateForAT(p->b.movestat.exp, TRUE);
      at = evaluateForAT(p->b.movestat.loc, TRUE);
      break;
    case P_BLOCK:
      doTypeCheck(p->b.block.stats);
      doTypeCheck(p->b.block.unavailablehandler);
      doTypeCheck(p->b.block.failurehandler);
      break;
    case P_PRIMSTAT:
      checkPrimitiveStatement(p);
      return;
    case P_UNAVAILABLEHANDLER:
      doTypeCheck(p->b.unavailablehandler.decl);
      doTypeCheck(p->b.unavailablehandler.body);
      break;
    case P_WAITSTAT:
      q = p->b.waitstat.exp;
      at = evaluateForAT(q, TRUE);
      if (! conforms(at, refToBuiltin(B_INSTAT, CONDITIONINDEX))) {
	ErrorMessage(q, "Operand of wait must be a Condition");
      }
      break;
    case P_SIGNALSTAT:
      q = p->b.signalstat.exp;
      at = evaluateForAT(q, TRUE);
      if (! conforms(at, refToBuiltin(B_INSTAT, CONDITIONINDEX))) {
	ErrorMessage(q, "Operand of signal must be a Condition");
      }
      break;
    default:
      fprintf(stderr, "Bad tag (%s) in typeCheck.\n", 
	tagNames[(int)p->tag]);
      break;
  }
}

static Boolean isWritable(kind)
ST_Kinds kind;
{
  return (kind == ST_Var || kind == ST_Result);
}

static void checkPrimitiveStatement(p)
NodePtr p;
{
  /*
   * This one needs to have access to the primitive descriptors, and check
   * that the references are appropriate with respect to read/write
   * permission, right number, and right concrete types.
   */
  NodePtr syms;
  Symbol st[3];
  register PrimitiveDescPtr pdp;
  register int pNum, i, j;

  assert(p->tag == P_PRIMSTAT);
  if (p->b.primstat.number->tag == P_STRINGLIT) {
    /*
     * This is a C implemented primitive, and we do not know enough to check
     * it.  Trust me!!
     */
  } else if (p->b.primstat.number->tag == P_INTLIT) {
    pNum = atoi(p->b.primstat.number->b.intlit.string);
    for (pdp = primitives; pdp->number != 0 && pdp->number != pNum; pdp++) ;
    if (pdp->number == 0) {
      ErrorMessage(p, "Unknown primitive number %d", pNum);
      return;
    }
    syms = p->b.primstat.vars;
    i = 0;
    if (Sequence_Length(syms) > 0) {
      if (! pdp->isFunction) {
	ErrorMessage(syms, "Primitive %d does not return a result", pNum);
      } else {
	if (Sequence_Length(syms) > 1) {
	  ErrorMessage(syms, "Primitive %d returns only one result", pNum);
	}
	st[0] = ST_Fetch(syms->b.children[0]->b.symref.symbol);
	if (! isWritable(st[i]->itsKind)) {
	  ErrorMessage(p, "The result of primitive %d must be writable", pNum);
	}
	if ((int)pdp->atindex[i] != ANYINDEX &&
	    !conforms(refToBuiltin(B_INSTAT, (int)pdp->atindex[i]),
	      st[i]->value.ATinfo)){
	  ErrorMessage(p, "The result primitive %d does not conform", pNum);
	}
      }
      i ++;
    }
    syms = p->b.primstat.vals;
    j = 0;
    while (i < 3) {
      if (pdp->atindex[i] != NILINDEX) {
	if (Sequence_Length(syms) <= j) {
	  ErrorMessage(p, "Primitive %d expects a %d%s parameter", pNum, j+1,
	    j == 0 ? "st" : j == 1 ? "nd" : "rd");
	} else {
	  st[i] = ST_Fetch(syms->b.children[j]->b.symref.symbol);
	  if (!conforms(st[i]->value.ATinfo,
	    refToBuiltin(B_INSTAT, (int)pdp->atindex[i]))) {
	    ErrorMessage(p, "Primitive %d%s argument does not conform",
	      j+1, j == 0 ? "st" : j == 1 ? "nd" : "rd");
	  }
	}
      } else if (Sequence_Length(syms) > j) {
	ErrorMessage(p, "Too many parameters to primitive %d", pNum);
      }
      i ++;
      j ++;
    }
    if (Sequence_Length(syms) > 3) {
      ErrorMessage(p, "Too many parameters to primitive %d", pNum);
    }
  }
}

static void checkAssignmentStatement(p)
NodePtr p;
{
  NodePtr left, right, q, aright;
  register NodePtr at;
  int i;
  Symbol st, lst;

  assert(p->tag == P_ASSIGNSTAT);
  left = p->b.assignstat.left;
  assert((int) p->b.assignstat.op == OASSIGN);
  right = p->b.assignstat.right;
  Sequence_For(q, left)
    assert (q->tag == P_SYMREF);
    st = ST_Fetch(q->b.symref.symbol);
    if (st->itsKind != ST_Var && st->itsKind != ST_Result) {
      ErrorMessage(p, "Assignment to non-variable");
    }
  Sequence_Next
  aright = right->b.children[0];
  at = evaluateForAT(aright, FALSE);
  assert(at != NULL);
  if (at->tag == T_SEQUENCE) {
    if (right->nChildren > 1) {
      ErrorMessage(p, "Illegal right hand side of multiple assignment, all except the first procedure invocation will be ignored");
    }
    if (Sequence_Length(left) != Sequence_Length(at)) {
      ErrorMessage(p, 
	"Wrong number of results in procedure invocation, expected %d found %d",
	Sequence_Length(at), Sequence_Length(left));
    }
    for (i = 0; i < Sequence_Length(left) && i < Sequence_Length(at); i++) {
      lst = ST_Fetch(left->b.children[i]->b.symref.symbol);
      if (! conforms(at->b.children[i], lst->value.ATinfo)) {
	ErrorMessage(p, "Result #%d does not conform to target", i+1);
      }
    }
  } else {
    if (Sequence_Length(left) != Sequence_Length(right)) {
      ErrorMessage(p, "Bad assignment, %d on left %d on right", 
	Sequence_Length(left), Sequence_Length(right));
    }
    for (i = 0; i < Sequence_Length(right); i++) {
      if (i > 0) at = evaluateForAT(right->b.children[i], TRUE);
      if (i < Sequence_Length(left)) {
	lst = ST_Fetch(left->b.children[i]->b.symref.symbol);
	assert(lst->value.ATinfo != NN);
	if (! conforms(at, lst->value.ATinfo)) {
	  ErrorMessage(p, "Expression #%d does not conform to target", i+1);
	}
      }
    }
  }
}

/*
 * The function conforms takes two abstract types
 * and determines whether they conform or not.  Both of the abstract types
 * must have been completely resolved.  That is, all the abstract types must
 * be pointers to P_ATLIT nodes, and none of them can be 
 * missing.
 */

/*
 * The algorithm is as follows:
 *  if we know that at1 conforms to at2, then return TRUE
 *  else 
 *	1.  assume that at1 @> at2.
 *	2.  for each operation (o2) in at2 
 *	    a.  if no identically named operation exitsts in at1 then retract
 *		the assumption and return false.
 *	    b.  if numparams or numresults do not match then reject the
 *		assumption and return FALSE.
 *	    c.  for each param(p1, p2), 
 *		* note that parameters must conform in the backwards
 *		   direction.
 *		i.  if p1 @> p2 then continue.
 *		ii. else reject the assumption and return FALSE.
 *	    c.  for each result(r1, r2), 
 *		i.  if r2 @> r1 then continue
 *		ii. else reject the assumption and return FALSE.
 *	    d.  return true (having proved our current assumption that 
 *		t1 @> t2)
 *
 * The algorithm for conformity between anything and a type variable is 
 * exactly the same, except that at the
 * assumption stage we identify at1 and at2, that is we say that they conform
 * in both directions.  When we reject the assumption we must remember to
 * invalidate in both directions.
 */

init_Conforms()
{
}

Boolean internalConforms(fAT1, fAT2)
NodePtr fAT1, fAT2;
{
  Boolean answer;
  int saveTraceConformFailure = traceconformfailure;
  traceconformfailure = 0;
  answer = conforms(fAT1, fAT2);
  traceconformfailure = saveTraceConformFailure;
  return(answer);
}

Boolean conforms();

void checkConforms(fAT1, fAT2, fNodePtr)
NodePtr fAT1, fAT2;
NodePtr fNodePtr;
{
  if (!conforms(fAT1, fAT2)) {
    BeginErrorMessage(fNodePtr);
    ErrorWrite("Types do not conform");
    EndErrorMessage();
  }
}

typedef enum { IDENTICAL, CONFORMS, NOTCONFORMS } Relationship;

/* return the op name of a signature */
#define ONOF(p) ((p)->b.opsig.name->b.opname.id)
#define POF(p) ((p)->b.opsig.params)
#define ROF(p) ((p)->b.opsig.results)
#define FOF(p) ((p)->b.opsig.isFunction)

static Boolean relationshipResults [] = { TRUE, TRUE, FALSE };

static Map conformsMap = NULL;
static OID cacheKey = NULL;
static Map cacheMap;

static int guessLevel = 0;
typedef struct GuessRecord {
  int guessLevel;
  OID oid1, oid2;
  Relationship relationship;
  struct GuessRecord *prev;
} GuessRecord, *GuessRecordPtr;
static GuessRecordPtr guessRecords = NULL;

static Relationship getRelationship(oid1, oid2)
OID oid1, oid2;
{
  if (conformsMap == NULL) {
    conformsMap = Map_Create();
  }
  if (oid1 != cacheKey) {
    cacheKey = oid1;
    cacheMap = (Map) Map_Lookup(conformsMap, (int)oid1);
    if (cacheMap == (Map)NIL) {
      cacheMap = Map_Create();
      Map_Insert(conformsMap, (int)oid1, (int)cacheMap);
    }
  }
  return((Relationship)Map_Lookup(cacheMap, (int)oid2));
}

static void setRelationship(oid1, oid2, relationship)
OID oid1, oid2;
Relationship relationship;
{
  register GuessRecordPtr g;

  if (conformsMap == NULL) {
    conformsMap = Map_Create();
  }
  if (oid1 != cacheKey) {
    cacheKey = oid1;
    cacheMap = (Map) Map_Lookup(conformsMap, (int)oid1);
    if (cacheMap == (Map)NIL) {
      cacheMap = Map_Create();
      Map_Insert(conformsMap, (int)oid1, (int)cacheMap);
    }
  }
  Map_Insert(cacheMap, (int)oid2, (int)relationship);
  if (guessLevel > 0) {
    g = (GuessRecordPtr) malloc(sizeof(GuessRecord));
    g->guessLevel = guessLevel;
    g->oid1 = oid1;
    g->oid2 = oid2;
    g->relationship = relationship;
    g->prev = guessRecords;
    guessRecords = g;
  }
}
    
static void startGuessing(oid1, oid2, relationship)
OID oid1, oid2;
Relationship relationship;
{
  guessLevel++;
  setRelationship(oid1, oid2, relationship);
  if (relationship == IDENTICAL) setRelationship(oid2, oid1, IDENTICAL);
}

static void endGuessing(success)
Boolean success;
{
  register GuessRecordPtr g;
  for (g = guessRecords; g != NULL && g->guessLevel == guessLevel; g = g->prev) {
    if (!success) {
      setRelationship(g->oid1, g->oid2, (Relationship)NIL);
    }
    guessRecords = g->prev;
    free((char *)g);
  }
  guessLevel--;
}

static Boolean allConform(), hardConforms(), i_conforms();

Boolean conforms(at1, at2)
NodePtr at1, at2;
{
  Boolean result;
  result = i_conforms(at1, at2, 1);
  return(result);
}

static Boolean i_conforms(fAT1, fAT2, level)
NodePtr fAT1, fAT2;
int level;
{
  OID oid1, oid2;
  Relationship relationship;
  Boolean result;
  register NodePtr at1, at2;

  if (loadedDummyBuiltins || bflag) return(TRUE);
  at1 = figureOutAT(fAT1);
  at2 = figureOutAT(fAT2);

  assert (at1->tag == P_ATLIT);
  assert (at2->tag == P_ATLIT);

  if (at1->b.atlit.f.isTypeVariable || at2->b.atlit.f.isTypeVariable) {
    return(matches(at1, at2, level));
  }

  TRACE2(conform, level, "Conforming %s to %s", ATName(at1), ATName(at2));

  oid1 = at1->b.atlit.id;
  oid2 = at2->b.atlit.id;
  if (oid1 == oid2) {
    TRACE0(conform, level, "Same oid => true");
    return(TRUE);
  }
  if (oid1 == OIDOfBuiltin(B_INSTAT, NILINDEX)) {
    TRACE0(conform, level, "Nil conforms to anything => true");
    return(TRUE);
  } else if (oid2 == OIDOfBuiltin(B_INSTAT, NILINDEX)) {
    TRACE0(conform, level, "Only nil conforms to None => false");
    return(FALSE);
  }
  relationship = getRelationship(oid1, oid2);
  if ((int) relationship != NIL) {
    result = relationshipResults[(int)relationship];
    TRACE1(conform, level, "Already know => %s", result ? "true" : "false");
    if (!result) TRACE2(conformfailure, level + 3,
      "%s *> %s: We already know that it does not", NameOf(at1), NameOf(at2));
    return(result);
  }

  /* if at2 will not admit conformers, then give up */
  if (at2->b.atlit.f.cannotBeConformedTo) {
    if (at2->b.atlit.f.isVector && at1->b.atlit.f.isVector) {
      TRACE2(conform, level,
	"%s cannotBeConformedTo but %s is also a vector so keep going", 
	ATName(at2), ATName(at1));
    } else {
      TRACE1(conform, level, "%s cannotBeConformedTo => false", ATName(at2));
      TRACE3(conformfailure, level + 3,
	"%s *> %s: %s is primitive and cannot be conformed to", NameOf(at1),
	NameOf(at2), NameOf(at2));
      return(FALSE);
    }
  }

  startGuessing(oid1, oid2, CONFORMS);
  if (hardConforms(at1, at2, level+1)) {
    endGuessing(TRUE);
    result = TRUE;
    TRACE0(conform, level, "=> true");
  } else {
    endGuessing(FALSE);
    result = FALSE;
    TRACE0(conform, level, "=> false");
    TRACE2(conformfailure, level + 3, "%s *> %s: Therefore fails", NameOf(at1),
      NameOf(at2));
  }
  return(result);
}

static Boolean hardConforms(at1, at2, level)
NodePtr at1, at2;
int level;
{
  int numOps1, numOps2, i, j;		
  NodePtr ops1, ops2, sig1, sig2;

  /* check immutability */
  if (at2->b.atlit.f.immutable && !at1->b.atlit.f.immutable) {
    TRACE2(conform, level, "%s is immutable, %s isnt => false", ATName(at2),
      ATName(at1));
    TRACE2(conformfailure, level + 3, "%s *> %s: immutability mismatch", NameOf(at1),
      NameOf(at2));
    return(FALSE);
  }
  j = 0;
  ops1 = at1->b.atlit.ops;
  numOps1 = Sequence_Length(ops1);
  ops2 = at2->b.atlit.ops;
  numOps2 = Sequence_Length(ops2);
  for (i = 0; i < numOps2; i++) {
    sig2 = ops2->b.children[i];
    TRACE1(conform, level, "Checking operation %s", SigName(sig2));
    while (j < numOps1 && ONOF(ops1->b.children[j]) != ONOF(sig2)) {
      assert(j >= (numOps1 - 1) ||
	ONOF(ops1->b.children[j]) < ONOF(ops1->b.children[j+1]));
      j ++;
    }
    if (j >= numOps1) {
      TRACE1(conform, level, "%s not defined => false", SigName(sig2));
      TRACE3(conformfailure, level + 3, "%s *> %s: operation %s is not defined",
	NameOf(at1), NameOf(at2), SigName(sig2));
      return (FALSE);
    }
    sig1 = ops1->b.children[j];
    if (Sequence_Length(POF(sig1)) != Sequence_Length(POF(sig2))) {
      TRACE3(conform, level,
	"operation %s argument number mismatch %d != %d => false",
	SigName(sig1), Sequence_Length(POF(sig1)), Sequence_Length(POF(sig2)));
      TRACE5(conformfailure, level + 3,
	"%s *> %s: operation %s: argument number mismatch %d != %d",
	NameOf(at1), NameOf(at2), SigName(sig1), Sequence_Length(POF(sig1)),
	Sequence_Length(POF(sig2)));
      return(FALSE);
    }
    if (Sequence_Length(ROF(sig1)) != Sequence_Length(ROF(sig2))) {
      TRACE3(conform, level,
	"operation %s result number mismatch %d != %d => false",
	SigName(sig1), Sequence_Length(ROF(sig1)), Sequence_Length(ROF(sig2)));
      TRACE5(conformfailure, level + 3,
	"%s *> %s: operation %s: result number mismatch %d != %d",
	NameOf(at1), NameOf(at2), SigName(sig1), Sequence_Length(ROF(sig1)),
	Sequence_Length(ROF(sig2)));
      return(FALSE);
    }
    if (FOF(sig2) && !FOF(sig1)) {
      TRACE1(conform, level,
	"operation %s function mismatch => false", SigName(sig1));
      TRACE3(conformfailure, level + 3,
	"%s *> %s: operation %s must be a function", NameOf(at1), NameOf(at2),
	SigName(sig1));
      return(FALSE);
    }
    if (! allConform(POF(sig2), POF(sig1), level+1)) {
      TRACE1(conform, level,
	"operation %s arguments mismatch => false", SigName(sig1));
      TRACE3(conformfailure, level + 3,
	"%s *> %s: operation %s argument types mismatch", NameOf(at1), NameOf(at2),
	SigName(sig1));
      return(FALSE);
    }
    if (! allConform(ROF(sig1), ROF(sig2), level+1)) {
      TRACE1(conform, level,
	"operation %s results mismatch  => false", SigName(sig1));
      TRACE3(conformfailure, level + 3,
	"%s *> %s: operation %s result types mismatch", NameOf(at1), NameOf(at2),
	SigName(sig1));
      return(FALSE);
    }
    assert(j >= (numOps1 - 1) ||
      ONOF(ops1->b.children[j]) < ONOF(ops1->b.children[j+1]));
    j++;
    TRACE1(conform, level, "operation %s matches", SigName(sig1));
  }
  return(TRUE);
}

Boolean matches(fAT1, fAT2, level)
NodePtr fAT1, fAT2;
{
  NodePtr at1, at2;
  OID oid1, oid2;
  Relationship relationship;
  Boolean result;

  if (loadedDummyBuiltins || bflag) return(TRUE);
  at1 = figureOutAT(fAT1);
  at2 = figureOutAT(fAT2);

  assert (at1->tag == P_ATLIT);
  assert (at2->tag == P_ATLIT);

  TRACE2(conform, level, "Matching %s and %s", ATName(at1), ATName(at2));
  oid1 = at1->b.atlit.id;
  oid2 = at2->b.atlit.id;
  if (oid1 == oid2) {
    TRACE0(conform, level, "Same oid => true");
    return(TRUE);
  }
  if (oid1 == OIDOfBuiltin(B_INSTAT, NILINDEX)) {
    TRACE0(conform, level, "Nil conforms to anything => true");
    return(TRUE);
  }
  relationship = getRelationship(oid1, oid2);
  if ((int) relationship != NIL) {
    result = relationshipResults[(int)relationship];
    TRACE1(conform, level, "Already know => %s", result ? "true" : "false");
    if (!result) TRACE2(conformfailure, level + 3,
      "%s *> %s: We already know that it does not", NameOf(at1), NameOf(at2));
    return(result);
  }
  startGuessing(oid1, oid2, IDENTICAL);
  if (hardConforms(at1, at2, level+1)) {
    endGuessing(FALSE);
    result = TRUE;
    TRACE0(conform, level, "=> true");
  } else {
    endGuessing(FALSE);
    result = FALSE;
    TRACE2(conformfailure, level + 3, "%s *> %s: Therefore fails", NameOf(at1),
      NameOf(at2));
  }
  return(result);
}

static Boolean allConform(ps1, ps2, level)
NodePtr ps1, ps2;
int level;
{
  int i, limit;
  register NodePtr t1, t2;
  register Symbol st1, st2;
  limit = Sequence_Length(ps1);
  for (i = 0; i < limit; i++) {
    t1 = ps1->b.children[i];
    t2 = ps2->b.children[i];
    assert(t1->tag == P_PARAM);
    assert(t2->tag == P_PARAM);
    t1 = t1->b.param.sym;
    t2 = t2->b.param.sym;
    assert(t1->tag == P_SYMDEF);
    assert(t2->tag == P_SYMDEF);
    st1 = ST_Fetch(t1->b.symdef.symbol);
    st2 = ST_Fetch(t2->b.symdef.symbol);
    t1 = st1->value.ATinfo;
    t2 = st2->value.ATinfo;
    assert(t1 != NULL);
    assert(t2 != NULL);
    if (! i_conforms(t1, t2, level+1)) {
      TRACE1(conformfailure, level + 3, "Position %d: types do not conform", i+1);
      return(FALSE);
    }
  }
  return (TRUE);
}

extern Map manifestMap;

extern NodePtr newExecuteAsAT();

static NodePtr ATOfInvocation(/* fNodePtr, needExp */);
/* NodePtr fNodePtr; Boolean needExp; */

static NodePtr ATOfExpression(/* fNodePtr */);
/* NodePtr fNodePtr; */
static NodePtr ATOfUnaryExpression(/* fNodePtr */);
/* NodePtr fNodePtr; */
NodePtr ATOfVectorLiteral();

/*
 * This needs to generate an error message if it cannot figure out at least
 * the at of the expression.
 */
NodePtr evaluateForAT(fNodePtr, needExp)
register NodePtr fNodePtr;
Boolean needExp;
{
  register NodePtr q;
  NodePtr originalAT;
  NodePtr at = NULL;
  register Symbol st;
  int whichType;
  Value value;

  if (fNodePtr == NULL) {
    fprintf(stderr, "Null node in evaluateForAT.\n");
    assert(FALSE);
    return(NULL);
  } else if ((int) fNodePtr < 0x200) {
    fprintf(stderr, "Input token (probably) in evaluateForAT (%d) \"%s\".\n",
      (int) fNodePtr, TOKENNAME((Token)fNodePtr));
    assert(FALSE);
    return(NULL);
  }
  nextLineNumber = fNodePtr->lineNumber;
  switch (fNodePtr->tag) {
    case P_SYMREF:
      st = ST_Fetch(fNodePtr->b.symref.symbol);
      if (st->value.ATinfo == NULL) {
	ErrorMessage(fNodePtr, "Illegal forward reference to %s",
	  ST_SymbolName(st));
	at = refToBuiltin(B_INSTAT, NILINDEX);
      } else {
	at = st->value.ATinfo;
      }
      break;
    case P_GLOBALREF:
      resolveGlobal(fNodePtr, &value);
      at = value.ATinfo;
      break;
    case P_UNARYEXP:
      at = ATOfUnaryExpression(fNodePtr);
      break;
    case P_EXP:
      at = ATOfExpression(fNodePtr);
      break;
    case P_INVOC:
      at = ATOfInvocation(fNodePtr, needExp);
      break;
    case P_VIEW:
      /*
       * Return the same expression, with the type given by evaluating
       * the expression.
       */
      originalAT = evaluateForAT(fNodePtr->b.view.exp, TRUE);
      at = newExecuteAsAT(fNodePtr->b.view.type);
      /*
       * We can decide at this point whether we will need a runtime 
       * check or not, because if itsType conforms to itsTypesValue
       * (or the other way around) then we do not need the check.  TODO.
       */
      if (at != NULL) {
	if (conforms(originalAT, at)) {
	  WarningMessage(fNodePtr, "Unnecessary view expression");
	  fNodePtr->b.view.isRedundant = TRUE;
	}
      }
      break;
    case P_RESTRICT:
      /*
       * Return the same expression, with the type restricted to the
       * type of the type expression
       */
      originalAT = evaluateForAT(fNodePtr->b.restrict.exp, TRUE);
      at = newExecuteAsAT(fNodePtr->b.restrict.type);
      if (at != NULL) {
	checkConforms(originalAT, at, fNodePtr);
      }
      break;
    case P_ATLIT:
      at = refToBuiltin(B_INSTAT, SIGNATUREINDEX);
      if (checkingTypes) doTypeCheck(fNodePtr);
      break;
    case P_OBLIT:
      if (fNodePtr->b.oblit.myat == NN) {
	assert (! fNodePtr->b.oblit.f.typesAreAssigned);
	newAssignTypes(fNodePtr, 1);
      }
      if (checkingTypes) doTypeCheck(fNodePtr);
      at = fNodePtr->b.oblit.myat;
      break;
    case P_BUILTINLIT:
      whichType = fNodePtr->b.builtinlit.whichType;
      q = refToBuiltinFromToken(B_IT, whichType);
      at = q->b.oblit.myat;
      break;
    case P_STRINGLIT:
      at = refToBuiltin(B_INSTAT, STRINGINDEX);
      break;
    case P_CHARLIT:
      at = refToBuiltin(B_INSTAT, CHARACTERINDEX);
      break;
    case P_INTLIT:
      at = refToBuiltin(B_INSTAT, INTEGERINDEX);
      break;
    case P_REALLIT:
      at = refToBuiltin(B_INSTAT, REALINDEX);
      break;
    case P_BOOLLIT:
      at = refToBuiltin(B_INSTAT, BOOLEANINDEX);
      break;
    case P_NILLIT:
      at = refToBuiltin(B_INSTAT, NILINDEX);
      break;
    /*
     * These next three are a little bit tricky, since they do not have obvious
     * abstract types.
     */
    case P_SELFLIT:
      assert(FALSE);
      break;
    case P_VECTORLIT:
      at = ATOfVectorLiteral(fNodePtr);
      break;
    default:
      fprintf(stderr, "Something else in evaluateForAT, (%d) \"%s\".\n",
	(int)fNodePtr->tag, tagNames[(int)fNodePtr->tag]);
  }
  return(at);
}

NodePtr ATOfVectorLiteral(p)
register NodePtr p;
{
  NodePtr instType, arg, args, target, opName, invoc, q, aType;
  Value value;
  
  if (p->b.vectorlit.vectorType == NN) {
    if (p->b.vectorlit.type == NN) {
      /* we are supposed to figure it out */
      if (Sequence_Length(p->b.vectorlit.exp) == 0) {
	p->b.vectorlit.type = refToBuiltin(B_INSTAT, ANYINDEX);
      } else {
	p->b.vectorlit.type = evaluateForAT(p->b.vectorlit.exp->b.children[0], TRUE);
	Sequence_For(q, p->b.vectorlit.exp)
	  aType = evaluateForAT(q, TRUE);
	  if (! conforms(aType, p->b.vectorlit.type)) {
	    p->b.vectorlit.type = refToBuiltin(B_INSTAT, ANYINDEX);
	    break;
	  }
	Sequence_Next
      }
    } else {
      Sequence_For(q, p->b.vectorlit.exp)
	aType = evaluateForAT(q, TRUE);
	checkConforms(aType, p->b.vectorlit.type, q);
      Sequence_Next
    }
    instType = newExecuteAsAT(p->b.vectorlit.type);
    arg = Construct(P_ARG, 1, instType);
    args = Construct(T_SEQUENCE, 1, arg);
    target = Construct(P_BUILTINLIT, 0);
#ifdef MUTABLEVECTORLITERALS
    target->b.builtinlit.whichType = KVECTOR;
#else
    target->b.builtinlit.whichType = IMMUTABLEVECTORINDEX;
#endif
    opName = Construct(P_OPNAME, 0);
    opName->b.opname.ident = Ident_Lookup("of", 2);
    opName->b.opname.id = ON_Translate("of");
    invoc = Construct(P_INVOC, 3, target, opName, args);
    newEvaluateManifest(invoc, &value);
    /* the answer is value.value */
    assert(value.value != NULL);
    p->b.vectorlit.vectorType = value.value;
  }
  return(figureOutAT(p->b.vectorlit.vectorType));
}

OID ownNameOID = 0, ownTypeOID = 0;
NodePtr ownNameSig = NN, ownTypeSig = NN;
NodePtr ownNameOpDef = NN, ownTypeOpDef = NN;

void initializeOwnStuff()
{
  NodePtr r1, r2, r3, r4, r;
  Symbol ownNameRes, ownTypeRes;
  ownNameOID = ON_Translate("ownname");
  ownTypeOID = ON_Translate("owntype");

  r1 = Construct(P_SYMDEF, 0);
  r1->b.symdef.ident = Ident_Lookup("@__0", 4);
  ownNameRes = ST_Create(NN, r1->b.symdef.ident);
  r1->b.symdef.symbol = ownNameRes;
  r2 = Construct(P_BUILTINLIT, 0);
  r2->b.builtinlit.whichType = KSTRING;
  r = Construct(P_PARAM, 3, r1, r2, NN);
  r3 = Construct(T_SEQUENCE, 1, r);
  r2 = NN;
  r1 = Construct(P_OPNAME, 0);
  r1->b.opname.ident = Ident_Lookup("ownname", 7);
  r1->b.opname.id = ownNameOID;
  r4 = NN;
  r = Construct(P_OPSIG, 4, r1, r2, r3, r4);
  r->b.opsig.isFunction = TRUE;
  ownNameSig = r;
  newAssignTypes(ownNameSig, 1);

  r1 = Construct(P_SYMDEF, 0);
  r1->b.symdef.ident = Ident_Lookup("@__0", 4);
  ownTypeRes = ST_Create(NN, r1->b.symdef.ident);
  r1->b.symdef.symbol = ownTypeRes;
  r2 = Construct(P_BUILTINLIT, 0);
  r2->b.builtinlit.whichType = KSIGNATURE;
  r = Construct(P_PARAM, 3, r1, r2, NN);
  r3 = Construct(T_SEQUENCE, 1, r);
  r2 = NN;
  r1 = Construct(P_OPNAME, 0);
  r1->b.opname.ident = Ident_Lookup("owntype", 7);
  r1->b.opname.id = ownTypeOID;
  r4 = NN;
  r = Construct(P_OPSIG, 4, r1, r2, r3, r4);
  r->b.opsig.isFunction = TRUE;
  ownTypeSig = r;
  newAssignTypes(ownTypeSig, 1);
  
  r1 = Construct(P_INTLIT, 0);
  r1->b.intlit.string = "0199";
  r2 = Construct(P_SYMREF, 0);
  r2->b.symref.symbol = ownNameRes;
  r2 = Construct(T_SEQUENCE, 1, r2);
  r1 = Construct(P_PRIMSTAT, 3, r1, r2, NN);
  r1 = Construct(T_SEQUENCE, 1, r1);
  ownNameOpDef = Construct(P_OPDEF, 2, ownNameSig,
    Construct(P_BLOCK, 3, r1, NN, NN));
  ownNameOpDef->b.opdef.isInlineable = TRUE;
  ownNameOpDef->b.opdef.isUniversal  = TRUE;
  ownNameOpDef->b.opdef.opNumber = -1;

  r1 = Construct(P_INTLIT, 0);
  r1->b.intlit.string = "0299";
  r2 = Construct(P_SYMREF, 0);
  r2->b.symref.symbol = ownTypeRes;
  r2 = Construct(T_SEQUENCE, 1, r2);
  r1 = Construct(P_PRIMSTAT, 3, r1, r2, NN);
  r1 = Construct(T_SEQUENCE, 1, r1);
  ownTypeOpDef = Construct(P_OPDEF, 2, ownTypeSig,
      Construct(P_BLOCK, 3, r1, NN, NN));
  ownTypeOpDef->b.opdef.isInlineable = TRUE;
  ownTypeOpDef->b.opdef.isUniversal  = TRUE;
  ownTypeOpDef->b.opdef.opNumber = -2;
}

NodePtr findATSignature(at, opName, opNumberPtr)
NodePtr at, opName;
int *opNumberPtr;
{
  register NodePtr sigs, asig, aname;
  register OID theID;
  Boolean found = FALSE;
  
  assert(at->tag == P_ATLIT);
  assert(opName->tag == P_OPNAME);
  theID = opName->b.opname.id;
  TRACE2(atctsort, 2, "findATSignature: at %s opName %s",
    ATName(at), ON_Name(theID));
  sigs = at->b.atlit.ops;
  assert(isASequence(sigs));
  *opNumberPtr = 0;
  asig = NULL;
  Sequence_For(asig, sigs)
    assert(asig->tag == P_OPSIG);
    aname = asig->b.opsig.name;
    assert(aname->tag == P_OPNAME);
    if (aname->b.opname.id == theID) {
      *opNumberPtr = z__z;
      found = TRUE;
      break;
    }
  Sequence_Next
  if (! found) {
    /*
     * It is possible that this is an invocation of ownName or ownType.
     */
    if (theID == ownNameOID) {
      *opNumberPtr = -1;
      asig = ownNameSig;
      found = TRUE;
    } else if (theID == ownTypeOID) {
      *opNumberPtr = -2;
      asig = ownTypeSig;
      found = TRUE;
    }
  }
  TRACE1(atctsort, 2, " number = %d", *opNumberPtr);
  return(found ? asig : NULL);
}

static NodePtr ATOfExpression(p)
register NodePtr p;
{
  register Token operator = p->b.exp.op;
  NodePtr lat, rat;

  lat = evaluateForAT(p->b.exp.left, TRUE);
  rat = evaluateForAT(p->b.exp.right, TRUE);
  if (operator == KAND || operator == KOR) {
    if (! conforms(lat, refToBuiltin(B_INSTAT, BOOLEANINDEX)) || ! conforms(rat, refToBuiltin(B_INSTAT, BOOLEANINDEX))){
      ErrorMessage(p, "Arguments of \"and\" and \"or\" must be Boolean");
    }
  } else if (operator == OIDENTITY) {
  } else if (operator == ONOTIDENTITY) {
  } else if (operator == OCONFORMSTO) {
    if (! conforms(lat, refToBuiltin(B_INSTAT, ABSTRACTTYPEINDEX))) {
      ErrorMessage(p, "Left argument of \"*>\" must be an abstract type");
    }
    rat = newExecuteAsAT(p->b.exp.right);
    if (rat == NULL) {
      ErrorMessage(p, "Right argument of \"*>\" must be a manifest abstract type");
    }
  } else {
    assert(FALSE);
  }
  return(refToBuiltin(B_INSTAT, BOOLEANINDEX));
}

static NodePtr ATOfUnaryExpression(p)
register NodePtr p;
{
  register Token operator = p->b.unaryexp.op;
  NodePtr lat, at = NULL;

  lat = evaluateForAT(p->b.unaryexp.exp, TRUE);
  if (operator == KAWAITING) {
    if (! conforms(lat, refToBuiltin(B_INSTAT, CONDITIONINDEX))) {
      ErrorMessage(p, "Arguments of \"awaiting\" must be Conditions");
    }
    at = refToBuiltin(B_INSTAT, BOOLEANINDEX);
  } else if (operator == KLOCATE) {
    at = refToBuiltin(B_INSTAT, NODEINDEX);
  } else if (operator == KISFIXED) {
    at = refToBuiltin(B_INSTAT, BOOLEANINDEX);
  } else {
    assert(FALSE);
  }
  return(at);
}

OID OIDOf(p)
register NodePtr p;
{
  register NodePtr q, r;
  q = GETVALUE(p);
  if (q->tag == P_BUILTINLIT) {
    r = refToBuiltinFromToken(B_INSTAT, q->b.builtinlit.whichType);
  } else {
    r = q;
  }
  assert(r->tag == P_ATLIT || r->tag == P_OBLIT);
  if (r->b.atlit.id == 0) {
    assert(r->b.oblit.f.dependsOnTypeVariable ||
	   r->b.oblit.f.inExecutableConstruct);
  }
  return(r->b.atlit.id);
}

#define ISME(p) ((p)->tag==P_SYMREF && stack != NULL && (p)->b.symref.symbol==stack->objectSym)

static NodePtr ATOfInvocation(p, needExp)
NodePtr p;
Boolean needExp;
{
  NodePtr target, targetAT, opName, parameters, signature, arguments;
  NodePtr formal, actual, results, aresult, valueATs, at, answer;
  Boolean isExpression;
  Symbol st;
  int opNumber;
  Boolean isManifest = FALSE;

  assert(p->tag == P_INVOC);
  TRACE2(typecheck, 3, "ATOfInvocation (%s) on line %d", showInvoc(p),
    p->lineNumber);
  answer = (NodePtr) Map_Lookup(manifestMap, (int)p+2);
  if ((int)answer != NIL) {
    /*
     * The answer was determined at compile time.  If we are checking types,
     * typeCheck it.
     */
    isManifest = TRUE;
    if (checkingTypes) typeCheck(answer);
  }

  if (p->b.invoc.resultTypeOID == 0) {
    assert(!isManifest);
  } else {
    at = OTLookup(p->b.invoc.resultTypeOID);
    assert(at != NULL);
    TRACE0(typecheck, 3, "Invocation result is manifest");
    return(at);
  }
  assert((int)answer == NIL);

  target = p->b.invoc.target;
  opName = p->b.invoc.opname;
  arguments = p->b.invoc.args;
  targetAT = evaluateForAT(target, TRUE);
  if (targetAT == NULL) {
    if (loadedDummyBuiltins || bflag) {
      TRACE0(typecheck, 3, "Dummy builtins or bflag so forget it");
      return(refToBuiltin(B_ITSAT, INTEGERINDEX));
    } else {
      assert(FALSE);
    }
  }
  targetAT = figureOutAT(targetAT);
  signature = findATSignature(targetAT, opName, &opNumber);
  p->b.invoc.opNumber = opNumber;
  if (signature == NULL) {
    ErrorMessage(p, "Operation \"%s\" is not defined", 
      ON_Name(opName->b.opname.id));
    return(refToBuiltin(B_INSTAT, NILINDEX));
  }

  /*
   * Check for incestuous monitor calls.
   */
  if (ISME(target)) {
    NodePtr theop;
    theop = findObjectOperation(stack->object, opName);
    if (theop->b.opdef.isMonitored &&
	!theop->b.opdef.isPrivate && 
	stack->inMonitor) {
      ErrorMessage(p, "Call of monitored operation within monitor");
    } else if (theop->b.opdef.isPrivate && !stack->inMonitor) {
      ErrorMessage(p, "Call of private operation from outside monitor");
    }
  }

  isExpression = signature->b.opsig.isFunction;
  if (needExp && !isExpression) {
    ErrorMessage(p, "Invocation of non-functional operation (%s) in expression context", ON_Name(signature->b.opsig.name->b.opname.id));
    return(refToBuiltin(B_INSTAT, NILINDEX));
  }
  parameters = signature->b.opsig.params;
  if (Sequence_Length(arguments) != Sequence_Length(parameters)) {
    ErrorMessage(p, "Wrong number of arguments: expected %d found %d",
      Sequence_Length(parameters), Sequence_Length(arguments));
  }
  valueATs = F_NewNode(T_SEQUENCE, Sequence_Length(arguments));
  valueATs->nChildren += Sequence_Length(arguments);
  Sequence_For(actual, arguments)
    assert(actual->tag == P_ARG);
    actual = actual->b.arg.exp;
    valueATs->b.children[z__z] = evaluateForAT(actual, TRUE);
    if (z__z < Sequence_Length(parameters)) {
      formal = parameters->b.children[z__z];
      assert(formal->tag == P_PARAM);
      st = ST_Fetch(formal->b.param.sym->b.symdef.symbol);
      if (checkingTypes) {
	if (! conforms(valueATs->b.children[z__z], st->value.ATinfo)) {
	  ErrorMessage(actual, "Argument #%d does not conform to formal", z__z+1);
	}
	if (st->value.value != NULL) {
	  /* we need to do matching */
	  /* This should have been a manifest invocation. */
	  assert(FALSE);
	}
      }
    }
  Sequence_Next
  results = signature->b.opsig.results;
  if (isExpression) {
    aresult = results->b.children[0];
    assert(aresult->tag == P_PARAM);
    st = ST_Fetch(aresult->b.param.sym->b.symdef.symbol);
    at = GETVALUE(st->value.ATinfo);
    if (at == NN) {
      at = newExecuteAsAT(aresult->b.param.type);
    }
    assert(at != NN && at->tag == P_ATLIT);
    if (checkingTypes) p->b.invoc.resultTypeOID = OIDOf(at);
  } else {
    at = F_NewNode(T_SEQUENCE, Sequence_Length(results));
    at->nChildren = Sequence_Length(results);
    Sequence_For(aresult, results)
      assert(aresult->tag == P_PARAM);
      st = ST_Fetch(aresult->b.param.sym->b.symdef.symbol);
      at->b.children[z__z] = GETVALUE(st->value.ATinfo);
      if (at->b.children[z__z] == NN) {
	at->b.children[z__z] = newExecuteAsAT(aresult->b.param.type);
      }
      assert(at->b.children[z__z] != NN && at->b.children[z__z]->tag == P_ATLIT);
    Sequence_Next
    if (Sequence_Length(results) > 0)
      if (checkingTypes) p->b.invoc.resultTypeOID = OIDOf(at->b.children[0]);
  }
  TRACE1(typecheck, 3, "ATOfInvocation returning at (%s)", ATName(at));
  return(at);
}
